#
# Copyright (c) 2015-2020 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_zodb_browser.zmi.views module

"""

__docformat__ = 'restructuredtext'

import time
from html import escape

import transaction
from ZODB.POSException import POSKeyError
from ZODB.utils import oid_repr, p64, tid_repr, u64
from persistent import Persistent, TimeStamp
from persistent.timestamp import TimeStamp
from pyramid.httpexceptions import HTTPFound
from pyramid.view import view_config
from zope.exceptions import UserError
from zope.interface import Interface, implementer

from pyams_layer.interfaces import IPyAMSLayer
from pyams_pagelet.pagelet import pagelet_config
from pyams_security.interfaces.base import MANAGE_SYSTEM_PERMISSION
from pyams_site.interfaces import PYAMS_APPLICATION_DEFAULT_NAME, PYAMS_APPLICATION_SETTINGS_KEY
from pyams_skin.interfaces.view import IInnerPage
from pyams_template.template import template_config
from pyams_utils.adapter import ContextRequestAdapter
from pyams_utils.property import cached_property
from pyams_viewlet.viewlet import viewlet_config
from pyams_zmi.interfaces import IAdminLayer
from pyams_zmi.interfaces.viewlet import IControlPanelMenu
from pyams_zmi.zmi.viewlet.menu import NavigationMenuItem
from pyams_zodb_browser.diff import compare_dicts_html
from pyams_zodb_browser.history import ZODBObjectHistory
from pyams_zodb_browser.interfaces import IDatabaseHistory, IValueRenderer
from pyams_zodb_browser.state import ZODBObjectState


from pyams_zodb_browser import _
from pyams_zodb_browser.value import TRUNCATIONS, prune_truncations


def get_object_type(obj):
    """Object type getter"""
    cls = getattr(obj, '__class__', None)
    if type(obj) is not cls:
        return '%s - %s' % (type(obj), cls)
    return str(cls)


def get_object_type_short(obj):
    """Short object type getter"""
    cls = getattr(obj, '__class__', None)
    if type(obj) is not cls:
        return '%s - %s' % (type(obj).__name__, cls.__name__)
    return cls.__name__


def get_object_path(obj, tid):
    """Object path getter"""
    path = []
    seen_root = False
    state = ZODBObjectState(obj, tid)
    while True:
        if state.is_root():
            path.append('/')
            seen_root = True
        else:
            if path:
                path.append('/')
            if not state.get_name() and state.get_parent_state() is None:
                # not using hex() because we don't want L suffixes for
                # 64-bit values
                path.append('0x%x' % state.get_object_id())
                break
            path.append(state.get_name() or '???')
        state = state.get_parent_state()
        if state is None:
            if not seen_root:
                path.append('/')
                path.append('...')
                path.append('/')
            break
    return ''.join(path[::-1])


class ZODBObjectAttribute:
    """Object attributes"""

    def __init__(self, name, value, tid=None):
        self.name = name
        self.value = value
        self.tid = tid

    def rendered_name(self):
        """Attribute name renderer"""
        return IValueRenderer(self.name).render(self.tid)

    def rendered_value(self):
        """Attribute value renderer"""
        return IValueRenderer(self.value).render(self.tid)

    def __repr__(self):
        return '%s(%r, %r, %r)' % (self.__class__.__name__, self.name,
                                   self.value, self.tid)

    def __eq__(self, other):
        if self.__class__ != other.__class__:
            return False
        return (self.name, self.value, self.tid) == (other.name, other.value,
                                                     other.tid)

    def __ne__(self, other):
        return not self.__eq__(other)


@viewlet_config(name='zodbbrowser.menu', context=Interface, layer=IAdminLayer, 
                manager=IControlPanelMenu, weight=999,
                permission=MANAGE_SYSTEM_PERMISSION)
class ZODBBrowserMenu(NavigationMenuItem):
    """ZODB browser menu"""

    label = _("ZODB browser")
    icon_class = 'fas fa-database'
    href = '#zodbbrowser'


class VeryCarefulView(ContextRequestAdapter):
    """Base ZODB view"""

    made_changes = False

    @cached_property
    def jar(self):
        try:
            return self.request.annotations['ZODB.interfaces.IConnection']
        except (KeyError, AttributeError):
            obj = self.find_closest_persistent()
            if obj is None:
                raise Exception("ZODB connection not available for this request")
            return obj._p_jar

    @property
    def readonly(self):
        return self.jar.isReadOnly()

    def find_closest_persistent(self):
        """Find closest persistent object"""
        obj = self.context
        while not isinstance(obj, Persistent):
            try:
                obj = obj.__parent__
            except AttributeError:
                return None
        return obj


@pagelet_config(name='zodbbrowser', context=Interface, layer=IPyAMSLayer, 
                permission=MANAGE_SYSTEM_PERMISSION)
@template_config(template='templates/zodbinfo.pt')
@implementer(IInnerPage)
class ZODBInfoView(VeryCarefulView):
    """ZODB info view"""

    def update(self):
        super(ZODBInfoView, self).update()
        prune_truncations()
        self.obj = self.select_object_to_view()
        # Not using IObjectHistory(self.obj) because LP#1185175
        self.history = ZODBObjectHistory(self.obj)
        self.latest = True
        if self.request.params.get('tid'):
            self.state = ZODBObjectState(self.obj,
                                         p64(int(self.request.params['tid'], 0)),
                                         _history=self.history)
            self.latest = False
        else:
            self.state = ZODBObjectState(self.obj, _history=self.history)

        if 'CANCEL' in self.request.params:
            raise self._redirect_to_self()

        if 'ROLLBACK' in self.request.params:
            rtid = p64(int(self.request.params['rtid'], 0))
            self.requestedState = self._tid_to_timestamp(rtid)
            if self.request.params.get('confirmed') == '1':
                self.history.rollback(rtid)
                transaction.get().note('Rollback to old state %s'
                                       % self.requestedState)
                self.made_changes = True
                transaction.get().commit()
                raise self._redirect_to_self()

    def _redirect_to_self(self):
        return HTTPFound(self.get_url())

    def select_object_to_view(self):
        """View context getter"""
        params = self.request.params
        obj = None
        if 'oid' not in params:
            obj = self.find_closest_persistent()
            # Sanity check: if we're running in standalone mode,
            # self.context is a Folder in the just-created MappingStorage,
            # which we're not interested in.
            if obj is not None and obj._p_jar is not self.jar:
                obj = None
        if obj is None:
            if 'oid' in params:
                try:
                    oid = int(params['oid'], 0)
                except ValueError:
                    raise UserError('OID is not an integer: %r' % params['oid'])
            else:
                oid = self.get_root_oid()
            try:
                obj = self.jar.get(p64(oid))
            except POSKeyError:
                raise UserError('There is no object with OID 0x%x' % oid)
        return obj

    def get_requested_tid(self):
        if 'tid' in self.request.params:
            return self.request.params['tid']
        return None

    def get_requested_tid_nice(self):
        if 'tid' in self.request.params:
            return self._tid_to_timestamp(p64(int(self.request.params['tid'], 0)))
        return None

    def get_object_id(self):
        return self.state.get_object_id()

    def get_object_id_hex(self):
        return '0x%x' % self.state.get_object_id()

    def get_object_type(self):
        return get_object_type(self.obj)

    def get_object_type_short(self):
        return get_object_type_short(self.obj)

    def get_state_tid(self):
        return u64(self.state.tid)

    def get_state_tid_nice(self):
        return self._tid_to_timestamp(self.state.tid)

    def get_pickle_size(self):
        return len(self.state.pickled_state)

    def get_root_oid(self):
        root = self.jar.root()
        try:
            settings = self.request.registry.settings
            root = root[settings.get(PYAMS_APPLICATION_SETTINGS_KEY, 
                                     PYAMS_APPLICATION_DEFAULT_NAME)]
        except KeyError:
            pass
        return u64(root._p_oid)

    def locate(self, path):
        not_found = object()  # marker

        # our current position
        #   partial -- path of the last _persistent_ object
        #   here -- path of the last object traversed
        #   oid -- oid of the last _persistent_ object
        #   obj -- last object traversed
        partial = here = '/'
        oid = self.get_root_oid()
        obj = self.jar.get(p64(oid))

        steps = path.split('/')

        if steps and steps[0]:
            # 0x1234/sub/path -> start traversal at oid 0x1234
            try:
                oid = int(steps[0], 0)
            except ValueError:
                pass
            else:
                partial = here = hex(oid)
                try:
                    obj = self.jar.get(p64(oid))
                except KeyError:
                    oid = self.get_root_oid()
                    return dict(error='Not found: %s' % steps[0],
                                partial_oid=oid,
                                partial_path='/',
                                partial_url=self.get_url(oid))
                steps = steps[1:]

        for step in steps:
            if not step:
                continue
            if not here.endswith('/'):
                here += '/'
            here += step.encode('utf-8')
            try:
                child = obj[step]
            except Exception:
                child = getattr(obj, step, not_found)
                if child is not_found:
                    return dict(error='Not found: %s' % here,
                                partial_oid=oid,
                                partial_path=partial,
                                partial_url=self.get_url(oid))
            obj = child
            if isinstance(obj, Persistent):
                partial = here
                oid = u64(obj._p_oid)
        if not isinstance(obj, Persistent):
            return dict(error='Not persistent: %s' % here,
                        partial_oid=oid,
                        partial_path=partial,
                        partial_url=self.get_url(oid))
        return dict(oid=oid,
                    url=self.get_url(oid))

    def get_url(self, oid=None, tid=None):
        if oid is None:
            oid = self.get_object_id()
        url = "#zodbbrowser?oid=0x%x" % oid
        if tid is None and 'tid' in self.request.params:
            url += "&tid=" + self.request.params['tid']
        elif tid is not None:
            url += "&tid=0x%x" % tid
        return url

    def get_breadcrumbs(self):
        breadcrumbs = []
        state = self.state
        seen_root = False
        while True:
            url = self.get_url(state.get_object_id())
            if state.is_root():
                breadcrumbs.append(('/', url))
                seen_root = True
            else:
                if breadcrumbs:
                    breadcrumbs.append(('/', None))
                if not state.get_name() and state.get_parent_state() is None:
                    # not using hex() because we don't want L suffixes for
                    # 64-bit values
                    breadcrumbs.append(('0x%x' % state.get_object_id(), url))
                    break
                breadcrumbs.append((state.get_name() or '???', url))
            state = state.get_parent_state()
            if state is None:
                if not seen_root:
                    url = self.get_url(self.get_root_oid())
                    breadcrumbs.append(('/', None))
                    breadcrumbs.append(('...', None))
                    breadcrumbs.append(('/', url))
                break
        return breadcrumbs[::-1]

    def get_path(self):
        return ''.join(name for name, url in self.get_breadcrumbs())

    def get_breadcrumbs_html(self):
        html = []
        for name, url in self.get_breadcrumbs():
            if url:
                html.append('<a href="%s">%s</a>' % (escape(url, True),
                                                     escape(name)))
            else:
                html.append(escape(name))
        return ''.join(html)

    def list_attributes(self):
        attrs = self.state.list_attributes()
        if attrs is None:
            return None
        return [ZODBObjectAttribute(name, value, self.state.requested_tid)
                for name, value in sorted(attrs)]

    def list_items(self):
        items = self.state.list_items()
        if items is None:
            return None
        return [ZODBObjectAttribute(name, value, self.state.requested_tid)
                for name, value in items]

    def _load_historical_state(self):
        results = []
        for d in self.history:
            try:
                interp = ZODBObjectState(self.obj, d['tid'],
                                         _history=self.history)
                state = interp.as_dict()
                error = interp.get_error()
            except Exception as e:
                state = {}
                error = '%s: %s' % (e.__class__.__name__, e)
            results.append(dict(state=state, error=error))
        results.append(dict(state={}, error=None))
        return results

    def list_history(self):
        """List transactions that modified a persistent object."""
        state = self._load_historical_state()
        results = []
        for n, d in enumerate(self.history):
            utc_timestamp = str(time.strftime('%Y-%m-%d %H:%M:%S',
                                              time.gmtime(d['time'])))
            local_timestamp = str(time.strftime('%Y-%m-%d %H:%M:%S',
                                                time.localtime(d['time'])))
            try:
                user_location, user_id = d['user_name'].split()
            except ValueError:
                user_location = None
                user_id = d['user_name']
            url = self.get_url(tid=u64(d['tid']))
            current = (d['tid'] == self.state.tid and
                       self.state.requested_tid is not None)
            curState = state[n]['state']
            oldState = state[n + 1]['state']
            diff = compare_dicts_html(curState, oldState, d['tid'])

            results.append(dict(utid=u64(d['tid']),
                                href=url, current=current,
                                error=state[n]['error'],
                                diff=diff, user_id=user_id,
                                user_location=user_location,
                                utc_timestamp=utc_timestamp,
                                local_timestamp=local_timestamp, **d))

        # number in reverse order
        for i in range(len(results)):
            results[i]['index'] = len(results) - i

        return results

    def _tid_to_timestamp(self, tid):
        if isinstance(tid, str) and len(tid) == 8:
            return str(TimeStamp(tid))
        return tid_repr(tid)


@view_config(name='zodbbrowser_path_to_oid', context=Interface, request_type=IPyAMSLayer,
             permission=MANAGE_SYSTEM_PERMISSION, renderer='json', xhr=True)
class PathToOidView(ZODBInfoView):

    def __call__(self):
        path = self.request.params.get('path')
        return self.locate(path)


@view_config(name='zodbbrowser_truncated', context=Interface, request_type=IPyAMSLayer,
             permission=MANAGE_SYSTEM_PERMISSION, renderer='json', xhr=True)
class TruncatedView(ZODBInfoView):

    def __call__(self):
        id = self.request.params.get('id')
        return TRUNCATIONS.get(id)


@pagelet_config(name='zodbbrowser_history', context=Interface, layer=IPyAMSLayer,
                permission=MANAGE_SYSTEM_PERMISSION)
@template_config(template='templates/zodbhistory.pt')
@implementer(IInnerPage)
class ZODBHistoryView(VeryCarefulView):
    """Zodb history view"""

    page_size = 5

    def update(self):
        super(ZODBHistoryView, self).update()
        prune_truncations()
        params = self.request.params
        if 'page_size' in params:
            self.page_size = max(1, int(params['page_size']))
        self.history = IDatabaseHistory(self.jar)
        if 'page' in params:
            self.page = int(params['page'])
        elif 'tid' in params:
            tid = int(params['tid'], 0)
            self.page = self.find_page(p64(tid))
        else:
            self.page = 0
        self.last_page = max(0, len(self.history) - 1) // self.page_size
        if self.page > self.last_page:
            self.page = self.last_page
        self.last_idx = max(0, len(self.history) - self.page * self.page_size)
        self.first_idx = max(0, self.last_idx - self.page_size)

    def get_url(self, tid=None):
        url = "#zodbbrowser_history"
        if tid is None and 'tid' in self.request.params:
            url += "?tid=" + self.request.params['tid']
        elif tid is not None:
            url += "?tid=0x%x" % tid
        return url

    def find_page(self, tid):
        try:
            pos = list(self.history.tids).index(tid)
        except ValueError:
            return 0
        else:
            return (len(self.history) - pos - 1) // self.page_size

    def list_history(self):
        if 'tid' in self.request.params:
            requested_tid = p64(int(self.request.params['tid'], 0))
        else:
            requested_tid = None

        results = []
        for n, d in enumerate(self.history[self.first_idx:self.last_idx]):
            utid = u64(d.tid)
            ts = TimeStamp(d.tid).timeTime()
            utc_timestamp = time.strftime('%Y-%m-%d %H:%M:%S', time.gmtime(ts))
            local_timestamp = time.strftime('%Y-%m-%d %H:%M:%S', time.localtime(ts))
            try:
                user_location, user_id = d.user.split()
            except ValueError:
                user_location = None
                user_id = d.user
            try:
                size = d._tend - d._tpos
            except AttributeError:
                size = None
            ext = d.extension if isinstance(d.extension, dict) else {}
            objects = []
            for record in d:
                obj = self.jar.get(record.oid)
                url = "#zodbbrowser?oid=0x%x&tid=0x%x" % (u64(record.oid),
                                                           utid)
                try:
                    objects.append(dict(
                        oid=u64(record.oid),
                        path=get_object_path(obj, d.tid),
                        oid_repr=oid_repr(record.oid),
                        class_repr=get_object_type(obj),
                        url=url,
                        repr=IValueRenderer(obj).render(d.tid),
                    ))
                except KeyError:  # no history
                    pass
            if len(objects) == 1:
                summary = '1 object record'
            else:
                summary = '%d object records' % len(objects)
            if size is not None:
                summary += ' (%d bytes)' % size
            results.append(dict(
                index=(self.first_idx + n + 1),
                utc_timestamp=utc_timestamp,
                local_timestamp=local_timestamp,
                user_id=user_id,
                user_location=user_location,
                description=d.description,
                utid=utid,
                current=(d.tid == requested_tid),
                href=self.get_url(tid=utid),
                size=size,
                summary=summary,
                hidden=(len(objects) > 5),
                objects=objects,
                **ext
            ))
        if results and not requested_tid and self.page == 0:
            results[-1]['current'] = True
        return results[::-1]



#
# Copyright (c) 2015-2020 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_*zodb_browser.value module

"""

import collections
import itertools
import logging
import re
from html import escape

from ZODB.utils import oid_repr, u64
from persistent import Persistent
from persistent.dict import PersistentDict
from persistent.list import PersistentList
from persistent.mapping import PersistentMapping
from zope.interface import Interface
from zope.interface.declarations import ProvidesClass

from pyams_utils.adapter import adapter_config
from pyams_zodb_browser.interfaces import IObjectHistory, IValueRenderer


__docformat__ = 'restructuredtext'


log = logging.getLogger(__name__)


MAX_CACHE_SIZE = 1000
TRUNCATIONS = {}
TRUNCATIONS_IN_ORDER = collections.deque()
next_id = itertools.count(1).__next__  # pylint: disable=invalid-name


def reset_truncations():  # for tests only!
    """Reset truncations"""
    global next_id  # pylint: disable=global-statement,invalid-name
    next_id = itertools.count(1).__next__
    TRUNCATIONS.clear()
    TRUNCATIONS_IN_ORDER.clear()


def prune_truncations():
    """Prune truncations"""
    while len(TRUNCATIONS_IN_ORDER) > MAX_CACHE_SIZE:
        del TRUNCATIONS[TRUNCATIONS_IN_ORDER.popleft()]


def truncate(text):
    """Truncate text"""
    id = 'tr%d' % next_id()  # pylint: disable=redefined-builtin,invalid-name
    TRUNCATIONS[id] = text
    TRUNCATIONS_IN_ORDER.append(id)
    return id


@adapter_config(required=Interface, provides=IValueRenderer)
class GenericValue:
    """Default value renderer.

    Uses the object's __repr__, truncating if too long.
    """

    def __init__(self, context):
        self.context = context

    def _repr(self):
        # hook for subclasses
        if getattr(self.context.__class__, '__repr__', None) is object.__repr__:
            # Special-case objects with the default __repr__ (LP#1087138)
            if isinstance(self.context, Persistent):
                return '<%s.%s with oid %s>' % (
                    self.context.__class__.__module__,
                    self.context.__class__.__name__,
                    oid_repr(self.context._p_oid))  # pylint: disable=protected-access
        try:
            return repr(self.context)
        except Exception:  # pylint: disable=broad-except
            try:
                return '<unrepresentable %s>' % self.context.__class__.__name__
            except Exception:  # pylint: disable=broad-except
                return '<unrepresentable>'

    def render(self, tid=None, can_link=True, limit=200):  # pylint: disable=unused-argument
        """Value renderer"""
        text = self._repr()
        if len(text) > limit:
            id = truncate(text[limit:])  # pylint: disable=redefined-builtin,invalid-name
            text = '%s<span id="%s" class="truncated">...</span>' % (
                escape(text[:limit]), id)
        else:
            text = escape(text)
        if not isinstance(self.context, str):
            try:
                n = len(self.context)  # pylint: disable=invalid-name
            except Exception:  # pylint: disable=broad-except
                pass
            else:
                if n == 1:  # this is a crime against i18n, but oh well
                    text += ' (%d item)' % n
                else:
                    text += ' (%d items)' % n
        return text


def join_with_commas(html, open, close):  # pylint: disable=redefined-builtin
    """Helper to join multiple html snippets into a struct."""
    prefix = open + '<span class="struct">'
    suffix = '</span>'
    for n, item in enumerate(html):  # pylint: disable=invalid-name
        if n == len(html) - 1:
            trailer = close
        else:
            trailer = ','
        if item.endswith(suffix):
            item = item[:-len(suffix)] + trailer + suffix
        else:
            item += trailer
        html[n] = item
    return prefix + '<br />'.join(html) + suffix


@adapter_config(required=str, provides=IValueRenderer)
class StringValue(GenericValue):
    """String renderer."""

    def __init__(self, context):  # pylint: disable=super-init-not-called
        self.context = context

    def render(self, tid=None, can_link=True, limit=200, threshold=4):  # pylint: disable=arguments-differ
        if self.context.count('\n') <= threshold:
            return GenericValue.render(self, tid, can_link=can_link,
                                       limit=limit)
        if isinstance(self.context, str):
            prefix = 'u'
            context = self.context
        else:
            prefix = ''
            context = self.context.decode('latin-1').encode('ascii', 'backslashreplace')
        lines = [re.sub(r'^[ \t]+',
                        lambda m: '&nbsp;' * len(m.group(0).expandtabs()),
                        escape(line))
                 for line in context.splitlines()]
        nl = '<br />'  # pylint: disable=invalid-name
        if sum(map(len, lines)) > limit:
            head = nl.join(lines[:5])
            tail = nl.join(lines[5:])
            id = truncate(tail)  # pylint: disable=redefined-builtin,invalid-name
            return (prefix + "'<span class=\"struct\">" + head + nl
                    + '<span id="%s" class="truncated">...</span>' % id
                    + "'</span>")
        return (prefix + "'<span class=\"struct\">" + nl.join(lines)
                + "'</span>")


@adapter_config(required=tuple, provides=IValueRenderer)
class TupleValue:
    """Tuple renderer."""

    def __init__(self, context):
        self.context = context

    def render(self, tid=None, can_link=True, threshold=100):
        """Value renderer"""
        html = []
        for item in self.context:
            html.append(IValueRenderer(item).render(tid, can_link))
        if len(html) == 1:
            html.append('')  # (item) -> (item, )
        result = '(%s)' % ', '.join(html)
        if len(result) > threshold or '<span class="struct">' in result:
            if len(html) == 2 and html[1] == '':
                return join_with_commas(html[:1], '(', ', )')
            return join_with_commas(html, '(', ')')
        return result


@adapter_config(required=list, provides=IValueRenderer)
class ListValue:
    """List renderer."""

    def __init__(self, context):
        self.context = context

    def render(self, tid=None, can_link=True, threshold=100):
        """Value renderer"""
        html = []
        for item in self.context:
            html.append(IValueRenderer(item).render(tid, can_link))
        result = '[%s]' % ', '.join(html)
        if len(result) > threshold or '<span class="struct">' in result:
            return join_with_commas(html, '[', ']')
        return result


@adapter_config(required=dict, provides=IValueRenderer)
class DictValue:
    """Dict renderer."""

    def __init__(self, context):
        self.context = context

    def render(self, tid=None, can_link=True, threshold=100):
        """Value renderer"""
        html = []
        for key, value in sorted(self.context.items()):
            html.append(IValueRenderer(key).render(tid, can_link) + ': ' +
                        IValueRenderer(value).render(tid, can_link))
        if (sum(map(len, html)) < threshold and
                '<span class="struct">' not in ''.join(html)):
            return '{%s}' % ', '.join(html)
        return join_with_commas(html, '{', '}')


@adapter_config(required=Persistent, provides=IValueRenderer)
class PersistentValue:
    """Persistent object renderer.

    Uses __repr__ and makes it a hyperlink to the actual object.
    """

    view_name = '#zodbbrowser'
    delegate_to = GenericValue

    def __init__(self, context):
        self.context = context

    def render(self, tid=None, can_link=True):
        """Value renderer"""
        obj = self.context
        url = '%s?oid=0x%x' % (self.view_name, u64(self.context._p_oid))  # pylint: disable=protected-access
        if tid is not None:
            url += "&tid=%d" % u64(tid)
            try:
                oldstate = IObjectHistory(self.context).load_state(tid)  # pylint: disable=assignment-from-no-return
                clone = self.context.__class__.__new__(self.context.__class__)
                clone.__setstate__(oldstate)
                clone._p_oid = self.context._p_oid  # pylint: disable=protected-access
                obj = clone
            except Exception:  # pylint: disable=broad-except
                log.debug('Could not load old state for %s 0x%x',
                          self.context.__class__, u64(self.context._p_oid))  # pylint: disable=protected-access
        value = self.delegate_to(obj).render(tid, can_link=False)
        if can_link:
            return '<a class="objlink" href="%s">%s</a>' % (escape(url), value)
        return value


@adapter_config(required=PersistentMapping, provides=IValueRenderer)
class PersistentMappingValue(PersistentValue):
    """Persistent mapping value adapter"""
    delegate_to = DictValue


@adapter_config(context=PersistentList, provides=IValueRenderer)
class PersistentListValue(PersistentValue):
    """Persistent list value adapter"""
    delegate_to = ListValue


if PersistentMapping is PersistentDict:
    # ZODB 3.9 deprecated PersistentDict and made it an alias for
    # PersistentMapping.  I don't know a clean way to conditionally disable the
    # <adapter> directive in ZCML to avoid conflicting configuration actions,
    # therefore I'll register a decoy adapter registered for a decoy class.
    # This adapter will never get used.

    class DecoyPersistentDict(PersistentMapping):
        """Decoy to avoid ZCML errors while supporting both ZODB 3.8 and 3.9."""

    @adapter_config(required=DecoyPersistentDict,
                    provides=IValueRenderer)
    class PersistentDictValue(PersistentValue):
        """Decoy to avoid ZCML errors while supporting both ZODB 3.8 and 3.9."""
        delegate_to = DictValue

else:
    @adapter_config(required=PersistentDict,
                    provides=IValueRenderer)
    class PersistentDictValue(PersistentValue):
        """Persistent dict value adapter"""
        delegate_to = DictValue


@adapter_config(required=ProvidesClass,
                provides=IValueRenderer)
class ProvidesValue(GenericValue):
    """zope.interface.Provides object renderer.

    The __repr__ of zope.interface.Provides is decidedly unhelpful.
    """

    def _repr(self):
        return '<Provides: %s>' % ', '.join(i.__identifier__
                                            for i in self.context._Provides__args[1:])  # pylint: disable=protected-access



#
# Copyright (c) 2015-2020 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_zodb_browser.btreesupport module

"""

__docformat__ = 'restructuredtext'

from BTrees.OOBTree import OOBTree, OOBucket
from zope.component import getMultiAdapter
from zope.container.btree import BTreeContainer
from zope.container.folder import Folder

from pyams_utils.adapter import adapter_config
from pyams_zodb_browser.history import ZODBObjectHistory
from pyams_zodb_browser.interfaces import IObjectHistory, IStateInterpreter
from pyams_zodb_browser.state import GenericState


@adapter_config(required=OOBTree, provides=IObjectHistory)
class OOBTreeHistory(ZODBObjectHistory):
    """OOBTree history adapter"""

    def _load(self):
        # find all objects (tree and buckets) that have ever participated in
        # this OOBTree
        queue = [self._obj]
        seen = set(self._oid)
        history_of = {}
        while queue:
            obj = queue.pop(0)
            history = history_of[obj._p_oid] = ZODBObjectHistory(obj)
            for d in history:
                state = history.load_state(d['tid'])
                if state and len(state) > 1:
                    bucket = state[1]
                    if bucket._p_oid not in seen:
                        queue.append(bucket)
                        seen.add(bucket._p_oid)
        # merge the histories of all objects
        by_tid = {}
        for h in list(history_of.values()):
            for d in h:
                by_tid.setdefault(d['tid'], d)
        self._history = list(by_tid.values())
        self._history.sort(key=lambda d: d['tid'], reverse=True)
        self._index_by_tid()

    def _last_real_change(self, tid=None):
        return ZODBObjectHistory(self._obj).last_change(tid)

    def load_state_pickle(self, tid=None):
        # lastChange would return the tid that modified self._obj or any
        # of its subobjects, thanks to the history merging done by _load.
        # We need the real last change value.
        # XXX: this is used to show the pickled size of an object.  It
        # will be misleading for BTrees if we show just the size for the
        # main BTree object while we're hiding all the individual buckets.
        return self._connection._storage.loadSerial(self._obj._p_oid,
                                                    self._last_real_change(tid))

    def load_state(self, tid=None):
        # lastChange would return the tid that modified self._obj or any
        # of its subobjects, thanks to the history merging done by _load.
        # We need the real last change value.
        return self._connection.oldstate(self._obj, self._last_real_change(tid))

    def rollback(self, tid):
        """Transaction rollback"""
        state = self.load_state(tid)
        if state != self.load_state():
            self._obj.__setstate__(state)
            self._obj._p_changed = True

        while state and len(state) > 1:
            bucket = state[1]
            bucket_history = IObjectHistory(bucket)
            state = bucket_history.load_state(tid)
            if state != bucket_history.load_state():
                bucket.__setstate__(state)
                bucket._p_changed = True


@adapter_config(required=(OOBTree, tuple, None),
                provides=IStateInterpreter)
class OOBTreeState:
    """Non-empty OOBTrees have a complicated tuple structure."""

    def __init__(self, type, state, tid):
        self.btree = OOBTree()
        self.btree.__setstate__(state)
        self.state = state
        # Large btrees have more than one bucket; we have to load old states
        # to all of them.  See BTreeTemplate.c and BucketTemplate.c for
        # docs of the pickled state format.
        while state and len(state) > 1:
            bucket = state[1]
            state = IObjectHistory(bucket).load_state(tid)
            # XXX this is dangerous!
            bucket.__setstate__(state)

        self._items = list(self.btree.items())
        self._dict = dict(self.btree)

        # now UNDO to avoid dangerous side effects,
        # see https://bugs.launchpad.net/zodbbrowser/+bug/487243
        state = self.state
        while state and len(state) > 1:
            bucket = state[1]
            state = IObjectHistory(bucket).load_state()
            bucket.__setstate__(state)

    def get_error(self):
        """Error getter"""
        return None

    def get_name(self):
        """Name getter"""
        return None

    def get_parent(self):
        """Parent getter"""
        return None

    def list_attributes(self):
        """Attributes getter"""
        return None

    def list_items(self):
        """Items getter"""
        return self._items

    def as_dict(self):
        """Dict representation"""
        return self._dict


@adapter_config(required=(OOBTree, type(None), None),
                provides=IStateInterpreter)
class EmptyOOBTreeState(OOBTreeState):
    """Empty OOBTrees pickle to None."""


@adapter_config(required=(Folder, dict, None),
                provides=IStateInterpreter)
class FolderState(GenericState):
    """Convenient access to a Folder's items"""

    def list_items(self):
        """Items getter"""
        data = self.state.get('data')
        if not data:
            return []
        # data will be an OOBTree
        loadedstate = IObjectHistory(data).load_state(self.tid)
        return getMultiAdapter((data, loadedstate, self.tid),
                               IStateInterpreter).list_items()


@adapter_config(required=(BTreeContainer, dict, None),
                provides=IStateInterpreter)
class BTreeContainerState(GenericState):
    """Convenient access to a BTreeContainer's items"""

    def list_items(self):
        # This is not a typo; BTreeContainer really uses
        # _SampleContainer__data, for BBB
        data = self.state.get('_SampleContainer__data')
        if not data:
            return []
        # data will be an OOBTree
        loadedstate = IObjectHistory(data).load_state(self.tid)
        return getMultiAdapter((data, loadedstate, self.tid),
                               IStateInterpreter).list_items()


@adapter_config(required=(OOBucket, tuple, None),
                provides=IStateInterpreter)
class OOBucketState(GenericState):
    """A single OOBTree bucket, should you wish to look at the internals

    Here's the state description direct from BTrees/BucketTemplate.c::

     * For a set bucket (self->values is NULL), a one-tuple or two-tuple.  The
     * first element is a tuple of keys, of length self->len.  The second element
     * is the next bucket, present if and only if next is non-NULL:
     *
     *     (
     *          (keys[0], keys[1], ..., keys[len-1]),
     *          <self->next iff non-NULL>
     *     )
     *
     * For a mapping bucket (self->values is not NULL), a one-tuple or two-tuple.
     * The first element is a tuple interleaving keys and values, of length
     * 2 * self->len.  The second element is the next bucket, present iff next is
     * non-NULL:
     *
     *     (
     *          (keys[0], values[0], keys[1], values[1], ...,
     *                               keys[len-1], values[len-1]),
     *          <self->next iff non-NULL>
     *     )

    OOBucket is a mapping bucket; OOSet is a set bucket.
    """

    def get_error(self):
        """Error getter"""
        return None

    def get_name(self):
        """Name getter"""
        return None

    def get_parent(self):
        """Parent getter"""
        return None

    def list_attributes(self):
        """Attributes getter"""
        return [('_next', self.state[1] if len(self.state) > 1 else None)]

    def list_items(self):
        """Items getter"""
        return list(zip(self.state[0][::2], self.state[0][1::2]))

    def as_dict(self):
        """Dict representation"""
        return dict(self.list_attributes(), _items=dict(self.list_items()))



#
# Copyright (c) 2015-2020 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_zodb_browser.state module

"""

import logging

import zope.interface.declarations
from ZODB.utils import u64
from persistent.dict import PersistentDict
from persistent.list import PersistentList
from persistent.mapping import PersistentMapping
from zope.component import getMultiAdapter
from zope.container.contained import ContainedProxy
from zope.container.ordered import OrderedContainer
from zope.container.sample import SampleContainer
from zope.interface import Interface, implementer
from zope.interface.interface import InterfaceClass
from zope.interface.interfaces import IInterface
from zope.traversing.interfaces import IContainmentRoot

from pyams_utils.adapter import adapter_config
from pyams_zodb_browser.interfaces import IObjectHistory, IStateInterpreter


__docformat__ = 'restructuredtext'


log = logging.getLogger(__name__)


real_Provides = zope.interface.declarations.Provides


def install_provides_hack():
    """Monkey-patch zope.interface.Provides with a more lenient version.

    A common result of missing modules in sys.path is that you cannot
    unpickle objects that have been marked with directlyProvides() to
    implement interfaces that aren't currently available.  Those interfaces
    are replaced by persistent broken placeholders, which aren classes,
    not interfaces, and aren't iterable, causing TypeErrors during unpickling.
    """
    zope.interface.declarations.Provides = Provides


def flatten_interfaces(args):
    """Flatten interfaces"""
    result = []
    for a in args:  # pylint: disable=invalid-name
        if isinstance(a, (list, tuple)):
            result.extend(flatten_interfaces(a))
        elif IInterface.providedBy(a):
            result.append(a)
        else:
            log.warning('  replacing %s with a placeholder', repr(a))
            result.append(InterfaceClass(a.__name__,
                                         __module__='broken ' + a.__module__))
    return result


def Provides(cls, *interfaces):  # pylint: disable=invalid-name
    """ZCA provides wrapper"""
    try:
        return real_Provides(cls, *interfaces)
    except TypeError as e:  # pylint: disable=invalid-name,broad-except
        log.warning('Suppressing TypeError while unpickling Provides: %s', e)
        args = flatten_interfaces(interfaces)
        return real_Provides(cls, *args)


@implementer(IStateInterpreter)
class ZODBObjectState:
    """Object state interpreter"""

    def __init__(self, obj, tid=None, _history=None):
        self.obj = obj
        if _history is None:
            _history = IObjectHistory(self.obj)
        else:
            assert _history._obj is self.obj
        self.history = _history
        self.tid = None
        self.requested_tid = tid
        self.load_error = None
        self.pickled_state = ''
        self._load()

    def _load(self):
        self.tid = self.history.last_change(self.requested_tid)
        try:
            self.pickled_state = self.history.load_state_pickle(self.tid)
            loaded_state = self.history.load_state(self.tid)
        except Exception as e:  # pylint: disable=broad-except,invalid-name
            self.load_error = "%s: %s" % (e.__class__.__name__, e)
            self.state = LoadErrorState(self.load_error, self.requested_tid)
        else:
            self.state = getMultiAdapter((self.obj, loaded_state, self.requested_tid),
                                         IStateInterpreter)

    def get_error(self):
        """Error getter"""
        return self.load_error

    def list_attributes(self):
        """Attributes getter"""
        return self.state.list_attributes()

    def list_items(self):
        """Items getter"""
        return self.state.list_items()

    def get_parent(self):
        """Parent getter"""
        return self.state.get_parent()

    def get_name(self):
        """Name getter"""
        name = self.state.get_name()
        if name is None:
            # __name__ is not in the pickled state, but it may be defined
            # via other means (e.g. class attributes, custom __getattr__ etc.)
            try:
                name = getattr(self.obj, '__name__', None)
            except Exception:
                # Ouch.  Oh well, we can't determine the name.
                pass
        return name

    def as_dict(self):
        """Dict representation"""
        return self.state.as_dict()

    # These are not part of IStateInterpreter

    def get_object_id(self):
        """Object ID getter"""
        return u64(self.obj._p_oid)

    def is_root(self):
        """Root checker"""
        return IContainmentRoot.providedBy(self.obj)

    def get_parent_state(self):
        """Parent state getter"""
        parent = self.get_parent()
        if parent is None:
            return None
        return ZODBObjectState(parent, self.requested_tid)


@implementer(IStateInterpreter)
class LoadErrorState:
    """Placeholder for when an object's state could not be loaded"""

    def __init__(self, error, tid):
        self.error = error
        self.tid = tid

    def get_error(self):
        """Error getter"""
        return self.error

    def get_name(self):
        """Name getter"""
        return None

    def get_parent(self):
        """Parent getter"""
        return None

    def list_attributes(self):
        """Attributes getter"""
        return []

    def list_items(self):
        """Items getter"""
        return None

    def as_dict(self):
        """Dict representation"""
        return {}


@adapter_config(required=(Interface, dict, None), provides=IStateInterpreter)
class GenericState:
    """Most persistent objects represent their state as a dict."""

    def __init__(self, type, state, tid):
        self.state = state
        self.tid = tid

    def get_error(self):
        """Error getter"""
        return None

    def get_name(self):
        """Name getter"""
        return self.state.get('__name__')

    def get_parent(self):
        """Parent getter"""
        return self.state.get('__parent__')

    def list_attributes(self):
        """Attributes getter"""
        return list(self.state.items())

    def list_items(self):
        """Items getter"""
        return None

    def as_dict(self):
        """Dict representation"""
        return self.state


@adapter_config(required=(PersistentMapping, dict, None),
                provides=IStateInterpreter)
class PersistentMappingState(GenericState):
    """Convenient access to a persistent mapping's items."""

    def list_items(self):
        """Items getter"""
        return sorted(self.state.get('data', {}).items())


if PersistentMapping is PersistentDict:
    # ZODB 3.9 deprecated PersistentDict and made it an alias for
    # PersistentMapping.  I don't know a clean way to conditionally disable the
    # <adapter> directive in ZCML to avoid conflicting configuration actions,
    # therefore I'll register a decoy adapter registered for a decoy class.
    # This adapter will never get used.

    class DecoyPersistentDict(PersistentMapping):
        """Decoy to avoid ZCML errors while supporting both ZODB 3.8 and 3.9."""

    @adapter_config(required=(DecoyPersistentDict, dict, None),
                    provides=IStateInterpreter)
    class PersistentDictState(PersistentMappingState):
        """Decoy to avoid ZCML errors while supporting both ZODB 3.8 and 3.9."""

else:

    @adapter_config(required=(PersistentDict, dict, None),
                    provides=IStateInterpreter)
    class PersistentDictState(PersistentMappingState):
        """Convenient access to a persistent dict's items."""


@adapter_config(required=(SampleContainer, dict, None),
                provides=IStateInterpreter)
class SampleContainerState(GenericState):
    """Convenient access to a SampleContainer's items"""

    def list_items(self):
        """Items getter"""
        data = self.state.get('_SampleContainer__data')
        if not data:
            return []
        # data will be something persistent, maybe a PersistentDict, maybe a
        # OOBTree -- SampleContainer itself uses a plain Python dict, but
        # subclasses are supposed to overwrite the _newContainerData() method
        # and use something persistent.
        loaded_state = IObjectHistory(data).load_state(self.tid)
        return getMultiAdapter((data, loaded_state, self.tid),
                               IStateInterpreter).list_items()


@adapter_config(required=(OrderedContainer, dict, None),
                provides=IStateInterpreter)
class OrderedContainerState(GenericState):
    """Convenient access to an OrderedContainer's items"""

    def list_items(self):
        # Now this is tricky: we want to construct a small object graph using
        # old state pickles without ever calling __setstate__ on a real
        # Persistent object, as _that_ would poison ZODB in-memory caches
        # in a nasty way (LP #487243).
        container = OrderedContainer()
        container.__setstate__(self.state)
        if isinstance(container._data, PersistentDict):
            old_data_state = IObjectHistory(container._data).load_state(self.tid)
            container._data = PersistentDict()
            container._data.__setstate__(old_data_state)
        if isinstance(container._order, PersistentList):
            old_order_state = IObjectHistory(container._order).load_state(self.tid)
            container._order = PersistentList()
            container._order.__setstate__(old_order_state)
        return list(container.items())


@adapter_config(required=(ContainedProxy, tuple, None), provides=IStateInterpreter)
class ContainedProxyState(GenericState):
    """Container proxy state interpreter"""

    def __init__(self, proxy, state, tid):
        GenericState.__init__(self, proxy, state, tid)
        self.proxy = proxy

    def get_name(self):
        """Name getter"""
        return self.state[1]

    def get_parent(self):
        """Parent getter"""
        return self.state[0]

    def list_attributes(self):
        """Attributes getter"""
        return [('__name__', self.get_name()),
                ('__parent__', self.get_parent()),
                ('proxied_object', self.proxy.__getnewargs__()[0])]

    def list_items(self):
        """Items getter"""
        return []

    def as_dict(self):
        """Dict representation"""
        return dict(self.list_attributes())


@adapter_config(required=(Interface, Interface, None),
                provides=IStateInterpreter)
class FallbackState:
    """Fallback when we've got no idea how to interpret the state"""

    def __init__(self, type, state, tid):
        self.state = state

    def get_error(self):
        """Error getter"""
        return None

    def get_name(self):
        """Name getter"""
        return None

    def get_parent(self):
        """Parent getter"""
        return None

    def list_attributes(self):
        """Attributes getter"""
        return [('pickled state', self.state)]

    def list_items(self):
        """Items getter"""
        return None

    def as_dict(self):
        """Dict representation"""
        return dict(self.list_attributes())



#
# Copyright (c) 2015-2020 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_zodb_browser.history module

"""

__docformat__ = 'restructuredtext'

import inspect

from ZODB.interfaces import IConnection
from ZODB.utils import tid_repr
from persistent import Persistent
from zope.interface import implementer

from pyams_utils.adapter import adapter_config
from pyams_zodb_browser.cache import get_storage_tids
from pyams_zodb_browser.interfaces import IDatabaseHistory, IObjectHistory


@adapter_config(required=Persistent, provides=IObjectHistory)
class ZODBObjectHistory:
    """ZODB object history adapter"""

    def __init__(self, obj):
        self._obj = obj
        self._connection = self._obj._p_jar
        self._storage = self._connection._storage
        self._oid = self._obj._p_oid
        self._history = None
        self._by_tid = {}

    def __len__(self):
        if self._history is None:
            self._load()
        return len(self._history)

    def _load(self):
        """Load history of changes made to a Persistent object.

        Returns a list of dictionaries, from latest revision to the oldest.
        The dicts have various interesting pieces of data, such as:

            tid -- transaction ID (a byte string, usually 8 bytes)
            time -- transaction timestamp (number of seconds since the Unix epoch)
            user_name -- name of the user responsible for the change
            description -- short description (often a URL)

        See the 'history' method of ZODB.interfaces.IStorage.
        """
        size = 999999999999  # "all of it"; ought to be sufficient
        # NB: ClientStorage violates the interface by calling the last
        # argument 'length' instead of 'size'.  To avoid problems we must
        # use positional argument syntax here.
        # NB: FileStorage in ZODB 3.8 has a mandatory second argument 'version'
        # FileStorage in ZODB 3.9 doesn't accept a 'version' argument at all.
        # This check is ugly, but I see no other options if I want to support
        # both ZODB versions :(
        if 'version' in inspect.getargspec(self._storage.history)[0]:
            version = None
            self._history = self._storage.history(self._oid, version, size)
        else:
            self._history = self._storage.history(self._oid, size=size)
        self._index_by_tid()

    def _index_by_tid(self):
        for record in self._history:
            self._by_tid[record['tid']] = record

    def __getitem__(self, item):
        if self._history is None:
            self._load()
        return self._history[item]

    def last_change(self, tid=None):
        """Get last change transaction id"""
        if self._history is None:
            self._load()
        if tid in self._by_tid:
            # optimization
            return tid
        # sadly ZODB has no API for get revision at or before tid, so
        # we have to find the exact tid
        for record in self._history:
            # we assume records are ordered by tid, newest to oldest
            if tid is None or record['tid'] <= tid:
                return record['tid']
        raise KeyError('%r did not exist in or before transaction %r' %
                       (self._obj, tid_repr(tid)))

    def load_state_pickle(self, tid=None):
        """Pickle state loader"""
        return self._connection._storage.loadSerial(self._obj._p_oid,
                                                    self.last_change(tid))

    def load_state(self, tid=None):
        """State loader"""
        return self._connection.oldstate(self._obj, self.last_change(tid))

    def rollback(self, tid):
        """Transaction rollback"""
        state = self.load_state(tid)
        if state != self.load_state():
            self._obj.__setstate__(state)
            self._obj._p_changed = True


@adapter_config(context=IConnection, provides=IDatabaseHistory)
@implementer(IObjectHistory)
class ZODBHistory:

    def __init__(self, connection):
        self._connection = connection
        self._storage = connection._storage
        self._tids = get_storage_tids(self._storage)

    @property
    def tids(self):
        return tuple(self._tids)  # readonlify

    def __len__(self):
        return len(self._tids)

    def __iter__(self):
        return self._storage.iterator()

    def __getitem__(self, index):
        if isinstance(index, slice):
            tids = self._tids[index]
            if not tids:
                return []
            return self._storage.iterator(tids[0], tids[-1])



#
# Copyright (c) 2015-2020 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_zodb_browser.cache module

"""

__docformat__ = 'restructuredtext'

import weakref
import time


MINUTES = 60
HOURS = 60 * MINUTES

STORAGE_TIDS = weakref.WeakKeyDictionary()


def expired(cache_dict, cache_for):
    """Expiration checker"""
    if 'last_update' not in cache_dict:
        return True
    return time.time() > cache_dict['last_update'] + cache_for


def get_storage_tids(storage, cache_for=5 * MINUTES):
    """Storage transactions ids getter"""
    cache_dict = STORAGE_TIDS.setdefault(storage, {})
    if expired(cache_dict, cache_for):
        if cache_dict.get('tids'):
            first = cache_dict['tids'][-1]
            last = cache_dict['tids'][-1]
            try:
                first_record = next(storage.iterator())
            except StopIteration:
                first_record = None
            if first_record and first_record.tid == first:
                # okay, look for new transactions appended at the end
                new = [t.tid for t in storage.iterator(start=last)]
                if new and new[0] == last:
                    del new[0]
                cache_dict['tids'].extend(new)
            else:
                # first record changed, we must've packed the DB
                cache_dict['tids'] = [t.tid for t in storage.iterator()]
        else:
            cache_dict['tids'] = [t.tid for t in storage.iterator()]
        cache_dict['last_update'] = time.time()
    return cache_dict['tids']
